<?php

class TrafficPreprocessWorker extends TemplateWorker
{
    protected $aseen = array();
    protected $traffic;
    protected $last_refresh = 0;
    protected $ports;
    protected $statements = array();
    protected $total_statements = 0;
    protected $datetime;
    protected $values_by_insert = 1000;
    protected $rows_by_gearman_connection = 1000;
    protected $max_if_id = 1000;

    public $runtimeRegister = array(
        'trafficpreprocess' => 'trafficpreprocess'
    );

    public function __construct()
    {
        parent::__construct();
        $this->traffic = new Traffic();
        $this->setValuesByInsert($this->values_by_insert);
        $this->setRowsByGearmanConnection($this->rows_by_gearman_connection);
    }

    public function trafficpreprocess($job)
    {
        $memory_data = json_decode(
            gzuncompress($job->workload()),
            true
        );

        $this->datetime = new Datetime();
        $do_refresh = (time() - $this->last_refresh > 30);
        if ($do_refresh) {
            $this->ports = $this->traffic->getPorts(null, true);
            $this->probes = $this->traffic->getProbes(true);
            $this->last_refresh = time();
        }

        try {
            $this->processRealtimeData($memory_data);
        } catch (Exception $e) {
            $this->debug("Exception caught:");
            $this->debug($e->getMessage());
        }

        unset($this->datetime);
    }

    /**
     * this function receive two ports (src and dst)
     * and set unregistered port as NULL. if both ports
     * are registered, function chooses the port with
     * minor value, the same is valid if none port is
     * registered
     *
     * @param string $port_src
     * @param string $port_dst
     */
    protected function selectPort($port_src, $port_dst)
    {

        // both ports are registered on database
        if ($this->ports[$port_src] && $this->ports[$port_dst]) {
            if ($port_src < $port_dst) {
                return $port_src;
            } else {
                return $port_dst;
            }
        }

        if ($this->ports[$port_src]) {
            return $port_src;
        }

        if ($this->ports[$port_dst]) {
            return $port_dst;
        }

        if ($port_src < $port_dst) {
            return $port_src;
        }

        return $port_dst;
    }

    /**
     * creates insert statement for aggregated
     * tables(realtime, hour, day and month) for
     * a given entry.
     *
     * parameters format follows:
     * <code>
     *     {
     *         "entry_time": "2015-02-06 16:57:03",
     *         "peer_ip_src": "192.168.10.1",
     *         "port": 53,
     *         "ip_proto": "udp",
     *         "iface_in": 1,
     *         "ip_src": "192.168.10.4",
     *         "packets": 1,
     *         "ip_dst": "192.168.10.1",
     *         "flows": 1,
     *         "bytes": 95
     *     }
     * </code>
     *
     * @param array $data
     */
    protected function createInsertStatements(&$data)
    {
        $this->total_statements++;

        $this->statements[] = sprintf(
            "('#DATETIME#','%s','%s','%s','%s','%s','%s','%s','%s','%s')",
            mysql_real_escape_string($data['peer_ip_src']),
            mysql_real_escape_string($data['port']),
            mysql_real_escape_string($data['iface_in']),
            mysql_real_escape_string($data['flows']),
            mysql_real_escape_string($data['ip_src']),
            mysql_real_escape_string($data['ip_dst']),
            mysql_real_escape_string($data['ip_proto']),
            mysql_real_escape_string($data['packets']),
            mysql_real_escape_string($data['bytes'])
        );
    }

    /**
     * This method define the number of rows/values in the same INSERT query.
     * Default value: 1000
     * @param Int $number_of_values
     */
    protected function setValuesByInsert($number_of_values)
    {
        if (!preg_match("/^[0-9]+$/", $number_of_values)) {
            throw new Exception("Invalid number of values by insert.");
        }

        $this->values_by_insert = $number_of_values;
    }

    /**
     * This method define the number of rows processed before sending jobs to gearman.
     * Default value: 1000
     * @param Int $number_of_rows
     */
    protected function setRowsByGearmanConnection($number_of_rows)
    {
        if (!preg_match("/^[0-9]+$/", $number_of_rows)) {
            throw new Exception("Invalid number of rows by gearman connection.");
        }

        $this->rows_by_gearman_connection = $number_of_rows;
    }

    /**
     * insert previously created statements
     * on gearman queue using multiple inserts
     * per query. the number of statements
     * per query is given by $this->values_by_insert
     */
    protected function insertStatements()
    {

        $query_tpl = "INSERT INTO traffic.aggregated_by_#TBL# (".
                     "    entry_time,".
                     "    peer_ip_src,".
                     "    port,".
                     "    iface_in,".
                     "    flows,".
                     "    ip_src,".
                     "    ip_dst,".
                     "    ip_proto,".
                     "    packets,".
                     "    bytes".
                     ") ".
                     "VALUES %s ".
                     "ON DUPLICATE KEY ".
                     "UPDATE bytes = bytes + VALUES(bytes), ".
                     "       packets = packets + VALUES(packets)";

        $st = array_splice($this->statements, 0, $this->values_by_insert);

        while (!empty($st)) {
            $realtime_q = str_replace("aggregated_by_#TBL#", "realtime", $query_tpl);
            $hour_q = str_replace("#TBL#", "hour", $query_tpl);
            $day_q = str_replace("#TBL#", "day", $query_tpl);

            $values = implode(",", $st);

            $realtime_q = sprintf(
                $realtime_q,
                str_replace(
                    "#DATETIME#",
                    $this->datetime->format('YmdHis'),
                    $values
                )
            );

            $hour_q = sprintf(
                $hour_q,
                str_replace(
                    "#DATETIME#",
                    $this->datetime->format('YmdH0000'),
                    $values
                )
            );

            $day_q = sprintf(
                $day_q,
                str_replace(
                    "#DATETIME#",
                    $this->datetime->format('Ymd000000'),
                    $values
                )
            );

            TaskRunner::runBackground('traffichistrealtime', gzcompress($realtime_q, 7));
            TaskRunner::runBackground('traffichisthour', gzcompress($hour_q, 7));
            TaskRunner::runBackground('traffichistday', gzcompress($day_q, 7));

            $st = array_splice($this->statements, 0, $this->values_by_insert);
        }

        unset($this->statements);
        $this->statements = array();
    }

    /**
     * Add new probes that are sending data to Traffic Analyzer.
     * @param Array $probes
     *     [
     *         {
     *             "<key>": {
     *                 "enabled": 0,
     *                 "probe": <peer_ip_src>,
     *                 "probe_desc": <peer_ip_src>,
     *                 "interface_id": <iface_in>,
     *                 "if_desc": <iface_in>
     *             }
     *     }
     *     ]
     */
    protected function addProbes(array $probes)
    {
        if (empty($probes)) {
            return true;
        }

        $query = "
			INSERT INTO opcfg.traffic_probes_interfaces (
				`enabled`,
				`probe`,
				`probe_desc`,
				`interface_id`,
				`if_desc`
			)
			VALUES
		";

        foreach ($probes as $key => $probe) {
            $query = sprintf(
                "%s (0,'%s','%s',%d,'%s'),",
                $query,
                $probe['probe'],
                $probe['probe_desc'],
                $probe['interface_id'],
                $probe['if_desc']
            );
        }

        $query = trim($query, ",");

        do_query($query, DBTRAFFIC);
    }

    /**
     * Update the last time that a probre sent data to Traffic Analyzer.
     * @param Array $probes
     *     [
     *         {
     *             "<key>": {
     *                 "probe": <peer_ip_src>,
     *                 "interface_id": <iface_in>,
     *                 "last_update": <date time on format Y-m-d H:i:s>
     *             }
     *     }
     *     ]
     */
    protected function updateProbes(array $probes)
    {

        if (empty($probes)) {
            return true;
        }

        foreach ($probes as $key => $probe) {
            $query = sprintf(
                "
			    UPDATE opcfg.traffic_probes_interfaces
			    SET last_update='%s'
			    WHERE probe='%s'
			    AND interface_id=%d ",
                $probe['last_update'],
                $probe['probe'],
                $probe['interface_id']
            );

            do_query($query, DBTRAFFIC);
        }
    }

    /**
     * Prepares the data for processing.
     *     - Remove invalid entries
     *     - Add new probes
     *     - Update all probes that sent data to Traffic Analyzer
     * @return array
     */
    protected function prepareMemoryData(&$memory_data)
    {
        if (!is_array($memory_data) || empty($memory_data)) {
            return array();
        }

        $entry_time = $this->datetime->format('Y-m-d H:i:s');
        $new_probes = array();
        $active_probes = array();

        $this->aseen = array();
        foreach ($memory_data as $idx => &$line) {
            try {
                Validator::validate_ip_address($line['peer_ip_src']);
                Validator::validate_ip_address($line['ip_src']);
                Validator::validate_ip_address($line['ip_dst']);
            } catch (InvalidArgumentException $e) {
                unset($memory_data[$idx]);
                continue;
            }

            if (!isset($line['bytes']) || $line['bytes'] <= 0) {
                unset($memory_data[$idx]);
                continue;
            }

            $tip = $line['peer_ip_src'];
            $tif = $line['iface_in'];
            $probe_key = sprintf(
                "%s_%s",
                $tip,
                $tif
            );
            if (!isset($this->probes[$tip][$tif])) {
                if (!$new_probes[$probe_key]) {
                    $new_probes[$probe_key] = array(
                        "enabled"   => 0,
                        "probe"         => $tip,
                        "probe_desc"    => $tip . '-' . $tif,
                        "interface_id"  => $tif,
                        "if_desc"   => $tif
                    );
                }
                continue;
            }

            if (!$active_probes[$probe_key]) {
                $active_probes[$probe_key] = array(
                    "probe"         => $tip,
                    "interface_id"  => $tif,
                    "last_update"   => $entry_time
                );
            }

            if (!$this->probes[$tip][$tif]['enabled']) {
                unset($memory_data[$idx]);
                continue;
            }

            $line['port'] = $this->selectPort(
                $line['port_src'],
                $line['port_dst']
            );
            unset($line['port_src']);
            unset($line['port_dst']);

            if ($this->aggregateFlow($line)) {
                unset($memory_data[$idx]);
            }

        }

        $this->addProbes($new_probes);
        $this->updateProbes($active_probes);
    }

    /**
     * Add up flows. If a given flow has already
     * been seen, we just aggregate the counters
     * on original flow and return true, otherwise
     * we keep the reference for the flow and
     * return false then on next time we are able
     * to aggregate.
     *
     * parameter format follows:
     * <code>
     *     [
     *         "entry_time": "2015-02-06 16:57:03",
     *         "peer_ip_src": "192.168.10.1",
     *         "port": 53,
     *         "ip_proto": "udp",
     *         "iface_in": 1,
     *         "ip_src": "192.168.10.4",
     *         "packets": 1,
     *         "ip_dst": "192.168.10.1",
     *         "flows": 1,
     *         "bytes": 95
     *     ]
     * </code>
     *
     * @param array $data
     * @return Boolean
     */
    protected function aggregateFlow(&$data)
    {

        $oidx = sprintf(
            "%s#%s#%s#%s#%s#%s",
            $data['peer_ip_src'],
            $data['port'],
            $data['iface_in'],
            $data['ip_src'],
            $data['ip_dst'],
            $data['ip_proto']
        );
        $idx = md5($oidx);
        if (isset($this->aseen[$idx])) {
            $this->aseen[$idx]['packets'] += $data['packets'];
            $this->aseen[$idx]['bytes'] += $data['bytes'];
            $this->aseen[$idx]['flows'] += $data['flows'];
            return true;
        }

        $this->aseen[$idx] = &$data;
        return false;
    }

    /**
     * process pmacct realtime data. here
     * we fetch data from pmacct process
     * and sum them up on databases
     */
    protected function processRealtimeData(&$memory_data)
    {
        // updates realtime data
        $this->prepareMemoryData($memory_data);

        foreach ($memory_data as &$line) {
            $pname = $line['peer_ip_src'];
            $ifid = $line['iface_in'];
            if ($ifid < 0 || $ifid > $this->max_if_id) {
                $fp = fopen("/tmp/strange-ports", "a");
                fprintf($fp, "%s\n", json_encode($line));
                fclose($fp);
                throw new Exception("Strange interface id has arrived: {$ifid}");
            }

            if (!isset($this->probes[$pname][$ifid]) ||
                !$this->probes[$pname][$ifid]) {
                continue;
            }

            $this->createInsertStatements($line);
            if ($this->total_statements > $this->rows_by_gearman_connection) {
                $this->insertStatements();
                $this->total_statements = 0;
            }
        }

        $this->insertStatements();
        unset($memory_data);
    }
}
