#!/usr/bin/perl
use strict;
use Data::Dumper;
use DBI;

our $time_to_die = 0;
our $server_config = undef;
our $dbconf = undef;;
our $logfile = "/var/log/opfprobed.log";
our $fprobepid = undef;
our $fprobebin = "/usr/sbin/fprobe";
our $dbconfile = "/usr/local/opmon/etc/db.php";
our $traffic_ip = undef;
our $traffic_port = undef;

#$| = 1;


sub sig_handler {
	$time_to_die = 1;
}

##
# parses opmon database configuration file
# and stores information on global variable
# $dbconf
#
sub parse_db_conf {

	return -1 if ( !-f $dbconfile);

	open(FILE,$dbconfile);
	foreach(<FILE>) {
		next if (!m/^ *\t*\$(.+) *=.*"(.*)".*;.*$/);
		$dbconf->{$1} = $2;
	}
	close(FILE);

	return 0;
}

##
# clone ourselves and start
# to monitor fprobe process
#
sub clone_and_dispatch {
	my $ret;
	my $counter = 0;
	my @pids;

	$ret = fork();
	return -1 if (!defined $ret);

	# parent ends here
	return 0 if ($ret != 0);

	$SIG{TERM} = \&sig_handler;
	$SIG{INT} = \&sig_handler;

	logit("fprobe keepalive forked");
	while(1) {
		@pids = locate_fprobe_pid();
		if (@pids == 0) {
			start_fprobe();
		} elsif (@pids > 1) {
			logit("multiple instances of fprobe running, killing...");
			foreach(@pids) {
				kill_pid($_);
			}
		} elsif (configuration_has_changed()) {
			kill_pid($pids[0]);
			start_fprobe();
		} elsif ($counter > 20) {
			logit("all up and running");
			$counter = 0;
		} else {
			$counter++;
		}

		if ($time_to_die) {
			logit("time to die has arrived. killing fprobes...");
			for(@pids) {
				kill_pid($_);
			}
			logit("done.");
			exit 0;
		}

		sleep(5);
	}

}

##
# returns 1 if pmacctd server configuration
# has changed, ip address or port
#
sub configuration_has_changed {
	my $newconf;

	$newconf = get_server_info();
	return 0 if (!defined $newconf);

	if ($server_config->{pmacctd_ip} == $newconf->{pmacctd_ip} &&
	    $server_config->{pmacctd_port} == $newconf->{pmacctd_port}) {
		return 0;
	}

	logit("configuration has changed:");
	logit("ip: ".$newconf->{pmacctd_ip});
	logit("port: ".$newconf->{pmacctd_port});

	$server_config = $newconf;
	return 1;
}

##
# retrieves server ip address
# and port
#
sub get_server_info {
	my $dsn;
	my $dbhandler;
	my $sth;
	my $row;
	my $newconf;

	return undef if (parse_db_conf() != 0);

	$dsn = sprintf(
	    "DBI:mysql:database=opcfg;host=%s;port=3306",
	    $dbconf->{DBHOST}
	);

    $dbhandler = DBI->connect(
        $dsn,
        $dbconf->{DBUSER},
        $dbconf->{DBPASS},
        {
            PrintError => 0
        }
    );

	if (!defined $dbhandler) {
		logit("unable to connect to mysql database");
		return undef;
	}

	$sth = $dbhandler->prepare(
	    "SELECT var,value ".
	    "FROM system_conf ".
	    "WHERE var='pmacctd_ip' ".
	    "OR var='pmacctd_port'"
	);
	if (!defined $sth->execute()) {
		logit("unable to run query on database");
		return undef;
	}

	if ($sth->rows != 2) {
		logit("error. query returned more than 2 rows");
		$sth->finish();
		$dbhandler->disconnect;
		return undef;
	}

	$row = $sth->fetchrow_hashref;
	$newconf->{$row->{var}} = $row->{value};
	$row = $sth->fetchrow_hashref;
	$newconf->{$row->{var}} = $row->{value};

	$sth->finish();
	$dbhandler->disconnect;

	if (!defined $newconf->{pmacctd_ip} ||
	    !defined $newconf->{pmacctd_port} ) {
		logit("error. no pmacctd_ip or pmacctd_port");
		return undef
	}

	return $newconf;
}

sub get_interfaces {
    my $interfaces;

    $interfaces = `ip link show | awk -F ': ' '/state UP/{print\$2}'`;

    chomp($interfaces);
    $interfaces =~ s/\n/,/g;

    return "any" if ($interfaces eq "");
    return $interfaces;
}

sub start_fprobe {
	my $cmd;
	logit("starting fprobe");

	$cmd = sprintf(
	    "%s -x 1 -i %s -d 1 -r 2 %s:%s",
	    $fprobebin,
        get_interfaces(),
	    $server_config->{pmacctd_ip},
	    $server_config->{pmacctd_port}
	);

	`$cmd`;
	logit("fprobe started");
	logit($cmd);
}

sub logit {
	my $now;
	my $msg = shift;

	$now = time();
	open(LOGFILE, ">>$logfile");
	print LOGFILE "[$now] $msg\n";
	close(LOGFILE);
}


sub kill_pid {
	my $pid = shift;
	my $alive = 1;
	my $max_tries = 10;
	my $try = 0;
	my $ret;

	logit("killing $pid");
	while($alive) {
		kill('SIGINT', $pid);
		$alive = kill(0, $pid);
		return 0 if (!$alive);

		sleep(1);
		$try++;
		if ($try == $max_tries) {

			logit("forced to kill $pid with SIGKILL");
			kill('SIGKILL', $pid);
			return 1;
		}
	}
}

sub locate_fprobe_pid {
	my @output;
	my @pids;
	my @cols;

	@output = `/bin/ps --no-headers -fC fprobe 2>/dev/null`;
	foreach(@output) {
		@cols = split(/\s+/, $_);
		next if ($#cols < 1);
		push(@pids, $cols[1]);
	}

	return @pids;
}

sub main {
	return 1 if (! -x $fprobebin);
	return 1 if (parse_db_conf() != 0);
	$server_config = get_server_info();
	return 1 if (!defined $server_config);
	return 1 if (clone_and_dispatch() != 0);
}

exit(main());
