<?php

require_once('/usr/local/opmon/etc/config.php');

class TrafficWriter {
    /**
     * This constant is a limit to maximum number of accepted flows.
     *
     * @const MAX_FLOWS Is a number of accepted flows before cut data.
     */
    const MAX_FLOWS = 1000;

    public $licensed;
    private $traffic;
    private $data;
    private $pipe;

    public function __construct()
    {
    	$this->traffic = new Traffic();
	set_logged_user("opmonadmin");
    	$lic = $this->traffic->checkLicense();
    	$this->licensed = $lic['licensed'];
    }

    /**
     * This method is used on usort function to select the most relevant flows.
     *
     * @param StdClass $a is a pmacct flow.
     * @param StdClass $b is a pmacct flow.
     *
     * @return Int
     */
    static public function sortCmpMethod($a, $b)
    {
        if ($a->bytes < $b->bytes) {
            return 1;
        }

        if ($a->bytes == $b->bytes) {
            return 0;
        }

        return -1;
    }

    /**
     * this function returns the memory data of
     * the OpMon Traffic Analyzer.
     *
     * Return format:
     * <code>
     * [
     *     {
     *         "peer_ip_src": "192.168.10.1",
     *         "port_dst": 53,
     *         "ip_proto": "udp",
     *         "iface_in": 1,
     *         "ip_src": "192.168.10.4",
     *         "packets": 1,
     *         "ip_dst": "192.168.10.1",
     *         "port_src": 37200,
     *         "flows": 1,
     *         "bytes": 95
     *     }
     * ]
     * </code>
     *
     * @return arr
     */
    protected function getMemoryData()
    {
    	$runner = new ShellCommandRunner();
    	$runner->run('/usr/bin/pmacct -s -O json -p ' . $this->pipe);

    	if ($runner->get_exit_code()) {
    		return array();
    	}

    	$tmp = str_replace("\n", ",", $runner->get_stdout());
    	unset($runner);

    	$this->data = json_decode(sprintf("[%s]", rtrim($tmp, ",")));
    }

    public function setPipe($file)
    {
    	if ( ! file_exists($file)) {
    		throw new InvalidArgumentException(
    			basename(__FILE__) . ": Invalid pipe file."
    		);
    	}

    	$this->pipe = $file;
    }

    /**
     * Send pmacct data to trafficpreprocess worker.
     */
    public function send()
    {
        set_time_limit(14);
    	if (is_null($this->data)) {
    		$this->getMemoryData();
        }

        if (count($this->data) > self::MAX_FLOWS) {
            usort($this->data, 'TrafficWriter::sortCmpMethod');
            $this->data = array_slice($this->data, 0, self::MAX_FLOWS);
        }

    	$chunks = array_chunk($this->data, 1000);
    	foreach ($chunks as $cks) {
    		$memdata = gzcompress(json_encode($cks), 7);
    		TaskRunner::runBackground("trafficpreprocess", $memdata);
    	}
    }
}

function help()
{
    print <<< EOL
    	This script is called to read netflow or sflow pipe and send data to
    	gearmand.


EOL;

    usage();
    print <<< EOL

    -h        Show this help.
    -p        Pipe file with full path.

EOL;
}

function usage()
{
    $file = basename(__FILE__);
    print <<< EOL
$file | -p <pipe file>

EOL;
}

function getOpts(Array &$opts)
{
    $opts = getopt("p:h");

    if (isset($opts['h'])) {
    	help();
    	exit(0);
    } elseif ( ! isset($opts['p'])) {
    	$opts['p'] = '/tmp/collect.pipe';
    }
}

try {
    $args = array();

    getOpts($args);

    $t = new TrafficWriter();

    if (!$t->licensed) {
    	exit(0);
    }

    $t->setPipe($args['p']);
    $t->send();
} catch (Exception $e) {
    print basename(__FILE__) . ": " . $e->getMessage() . PHP_EOL;
}

exit(0);
